@@ -54,6 +54,7 @@ class KeyframesDock;
 class MarkersDock;
 class NotesDock;
 class SubtitlesDock;
+class ScreenCapture;
 
 class MainWindow : public QMainWindow
 {
@@ -114,6 +115,7 @@ public:
 
 signals:
     void audioChannelsChanged();
+    void processingModeChanged();
     void producerOpened(bool withReopen = true);
     void profileChanged();
     void openFailed(QString);
@@ -160,6 +162,7 @@ private:
     bool checkAutoSave(QString &url);
     bool saveRepairedXmlFile(MltXmlChecker &checker, QString &fileName);
     void setAudioChannels(int channels);
+    void setProcessingMode(ShotcutSettings::ProcessingMode mode);
     void showSaveError();
     void setPreviewScale(int scale);
     void setVideoModeMenu();
@@ -171,6 +174,7 @@ private:
     void backupPeriodically();
     bool confirmProfileChange();
     bool confirmRestartExternalMonitor();
+    void resetFilterMenuIfNeeded();
 
     Ui::MainWindow *ui;
     Player *m_player;
@@ -215,6 +219,7 @@ private:
     SubtitlesDock *m_subtitlesDock;
     std::unique_ptr<QWidget> m_producerWidget;
     FilesDock *m_filesDock;
+    ScreenCapture *m_screenCapture;
 
 public slots:
     bool isCompatibleWithGpuMode(MltXmlChecker &checker);
@@ -302,7 +307,6 @@ private slots:
     void on_actionBicubic_triggered(bool checked);
     void on_actionHyper_triggered(bool checked);
     void on_actionJack_triggered(bool checked);
-    void on_actionGPU_triggered(bool checked);
     void onExternalTriggered(QAction *);
     void onDecklinkGammaTriggered(QAction *);
     void onKeyerTriggered(QAction *);
@@ -350,10 +354,8 @@ private slots:
     void on_actionAppDataSet_triggered();
     void on_actionAppDataShow_triggered();
     void on_actionNew_triggered();
-#if defined(Q_OS_MAC) || defined(Q_OS_WIN)
     void on_actionScreenSnapshot_triggered();
     void on_actionScreenRecording_triggered();
-#endif
     void on_actionKeyboardShortcuts_triggered();
     void on_actionLayoutLogging_triggered();
     void on_actionLayoutEditing_triggered();
